#!/bin/bash

shellfu import isa
shellfu import pretty
shellfu import sfpi

#
# Query on which distribution we are
#

#
# Override values for dist()
#
# Comma-separated list of KEY=VALUE pairs to prefer when
# querying for distribution meta-data.
#
# This mapping takes precedence before any actual querying
# and affects all querying functions, ie. dist(), dist__test()
# and dist__report().
#
# For dist__report() the list of known ids is derived by
# querying plugins plus this list, but any values found
# here will prevent further plugin querying.  For example:
#
#     DIST__OVERRIDE=id=foo,bar=baz dist__report
#
# will generate report with same amount of keys as normal
# report plus one extra key `bar`, and with value of `id`
# being `foo`.
#
# Derived values are affected by their main value.  For
# example, setting
#
#     DIST__OVERRIDE=ver=1.2.3
#
# affects not just `ver` but also `ver.x`, `ver.xy` and
# `ver.xyz`.
#
DIST__OVERRIDE=${DIST__OVERRIDE:=}

dist() {
    #
    # Get key $1 of distro info
    #
    # Usage:
    #
    #     dist KEY
    #
    # Find plugin supporting the key and use it
    #
    # Common KEYs are:
    #
    #  *  'id' - small-caps identifier, eg. `debian`,
    #  *  'name' - human-readable name, eg. `Red Hat Enterprise Linux`,
    #  *  'ver' - release version, eg. '8.1'
    #  *  'ver.x' - first part of rel. version, eg. '8'
    #  *  'ver.y' - second part of rel. version, eg. '1'
    #  *  'ver.z' - third part of rel. version, eg. '0'
    #
    # Few KEYs are supported from the Common Platform Enumeration (as
    # of version 2.3):
    #
    #  *  'cpe' - full CPE URI, eg. 'cpe:/o:fedoraproject:fedora:30',
    #  *  'cpe.vendor' - vendor, eg. 'fedoraproject',
    #  *  'cpe.product' - product name, eg. 'enterprise_linux',
    #  *  'cpe.version' - product version, eg. 'enterprise_linux',
    #  *  'cpe.update' - CPE update, eg. 'beta' (specific to distro).
    #
    # Note that distros don't have to support all parts of version; plugins
    # should report empty string if the part is not supported.  However,
    # since meaning is always distro-specific, you should not query for it
    # directly (eg. don't use `ver.z` if you are on Debian.).
    #
    local key=$1    # KEY of interest
    local plg       # each plugin name
    test -n "$key" || { mkusage -E KEY; return 2; }
    __dist__derive "$key" && return 0
    __dist__get_override "$key" && return 0
    #shellcheck disable=SC2034
    SFPI__PREFIX=_distp
    for plg in $(sfpi__ls); do
        debug -v plg
        sfpi__call "$plg" get "$key" \
          | grep . \
         && return
    done
    "$__DIST__WARNF" "no source for key: $key"
    return 3
}

dist__test() {
    #
    # Test on distro key
    #
    #     dist__test KEY OP VALUE [KEY OP VALUE]..
    #
    # True (exit with zero status) if VALUE matches queried value of KEY
    # according to comparison operator OP.  If multiple triplets of "KEY
    # OP VALUE" are provided, all comparisons must hold true, otherwise
    # the test fails (returns one).
    #
    # Known OPs are
    #
    #  *  'eq', 'lt', 'gt', 'le', 'ge' and 'ne' are integer comparison
    #     operators as known from `test` unix command.
    #
    #         dist__test ver.x gt 8     # true if version > 8
    #
    #  *  '==', '!=', are string comparison operators, also known from
    #     `test` unix command.
    #
    #         dist__test id == debian   # true on a Debian system
    #
    #  *  ',=' is true if value of KEY matches one of possible values
    #     in VALUE comma-separated list.
    #
    #         dist__test id ,= debian,ubuntu    # true on a Debian
    #                                           # .. or Ubuntu system
    #
    #  *  're' is regular expression matcher, ie. VALUE has to be basic
    #     regular expression matching whole line,
    #
    #         dist__test cpe re '.*fedora.*'    # true if CPE mentions
    #                                           # .. Fedora
    #
    #  *  'vge', 'vle`, `vgt` and `vlt` have similar meanings as their
    #     corresponding integer operators, but work on version numbers
    #     using 'sort -V' command.
    #
    #         dist__test ver.xy vge 1.13        # true if version number
    #                                           # .. is 1.13, and higher
    #
    # See dist() and dist__report() to find out about available KEYs.
    #
    while test -n "$1"; do
        __dist__test1 "$1" "$2" "$3" || return 1
        shift 3 || { mkusage -E "KEY OP VALUE [KEY OP VALUE].."; return 2; }
    done
    return 0
}

dist__report() {
    #
    # Report all known keys
    #
    #     dist__report
    #
    local key       # each key
    local plg       # each plugin name
    local val       # value obtained for each key
    local drver=0.2.5
    echo "# report generated by dist__report v$drver"
    #shellcheck disable=SC2034
    (
        __dist__enum_subs
        SFPI__PREFIX=_distp
        for plg in $(sfpi__ls); do
            debug -v plg
            sfpi__call "$plg" enum
        done
        echo "${DIST__OVERRIDE//,/$'\n'}" \
          | grep -Eo '[a-z]+[a-z.]*=' \
          | cut -d'=' -f1
    ) \
      | grep . \
      | sort \
      | uniq \
      | while read -r key; do
            val=$(__DIST__WARNF=true dist "$key")
            echo "$key=$val"
        done \
      | sort \
      | uniq
}

#
# Command to call when key source is missing
#
__DIST__WARNF=warn

__dist__enum_subs() {
    #
    # List of accessible compound keys
    #
    if dist cpe >/dev/null 2>/dev/null; then
        echo cpe.part
        echo cpe.vendor
        echo cpe.product
        echo cpe.version
        echo cpe.update
        echo cpe.edition
        echo cpe.lang
    fi
    if dist ver >/dev/null 2>/dev/null; then
        echo ver.x
        echo ver.xy
        echo ver.xyz
        echo ver.y
        echo ver.z
    fi
    if dist id >/dev/null 2>/dev/null; then
        echo id.vlike
    fi
}

__dist__derive() {
    #
    # Derive sub-key $1
    #
    # Key must be in form:
    #
    #   BASEKEY.SUBKEY
    #
    # and this function will obtain BASEKEY in normal
    # way and try to derive SUBKEY from it.
    #
    local key=$1
    case $key in
        cpe.*)  __dist__derive_cpe "$key" ;;
        ver.*)  __dist__derive_ver "$key" ;;
        id.*)   __dist__derive_id "$key" ;;
        *)      return 2 ;;
    esac
}

__dist__derive_cpe() {
    #
    # Derive 'cpe' sub-key $1
    #
    local key=$1
    local field
    local cpe
    cpe=$(dist cpe | grep '^cpe:/' | cut -d/ -f2)
    case $key in
        cpe.part)       field=1 ;;
        cpe.vendor)     field=2 ;;
        cpe.product)    field=3 ;;
        cpe.version)    field=4 ;;
        cpe.update)     field=5 ;;
        cpe.edition)    field=6 ;;
        cpe.lang)       field=7 ;;
        cpe.*)         return 2 ;;
    esac
    cut -d: -f"$field" <<<"$cpe"
}

__dist__derive_id() {
    #
    # Derive 'id' sub-key $1
    #
    local key=$1
    local id
    id=$(dist id)
    case $key:$id in
        id.vlike:rhel)    echo rhel ;;
        id.vlike:centos)  echo rhel ;;
        id.vlike:fedora)  echo fedora ;;
        id.vlike:debian)  echo debian ;;
        id.*)             return 2 ;;
    esac
}

__dist__derive_ver() {
    #
    # Derive 'ver' sub-key $1
    #
    local key=$1
    local ver
    ver=$(dist ver)
    case $key in
        ver.x)      __dist__derive_ver_take1 <<<"$ver" ;;
        ver.xy)     __dist__derive_ver_take2 <<<"$ver" ;;
        ver.xyz)    __dist__derive_ver_take3 <<<"$ver" ;;
        ver.y)      __dist__derive_ver ver.xy  | cut -d. -f2 ;;
        ver.z)      __dist__derive_ver ver.xyz | cut -d. -f3 ;;
        ver.*)      return 2 ;;
    esac
}

__dist__derive_ver_take1() {
    #
    # Take first version fragment
    #
    grep -o -E '^[[:digit:]]+'
}

__dist__derive_ver_take2() {
    #
    # Take first two version fragments
    #
    grep -o -E '^[[:digit:]]+[.][[:digit:]]+'
}

__dist__derive_ver_take3() {
    #
    # Take first three version fragments
    #
    grep -o -E '^[[:digit:]]+[.][[:digit:]]+[.][[:digit:]]+'
}

__dist__get_override() {
    #
    # Get key $1 override
    #
    local key=$1
    echo "${DIST__OVERRIDE//,/$'\n'}" \
      | grep -e "^${key//./'[.]'}=" \
      | cut -d'=' -f2 \
      | grep .
}

__dist__test1() {
    #
    # Test on distro key
    #
    #     dist__test1 KEY OP VALUE
    #
    local key=$1    # KEY of interest
    local Op=$2     # comparison OPerator
    local Ref=$3    # reference VALUE
    local Val       # real value
    test -n "$Op"
    Val=$(dist "$key") || return 3
    debug -v key Val Op Ref
    case $Op in
        eq|lt|le|gt|ge|ne)  __dist_intcmp             ;;
        ==)                 test "$Val" == "$Ref"     ;;
        !=)                 test "$Val" != "$Ref"     ;;
        re)                 grep -qxe "$Ref" <<<"$Val";;
        ,=)                 grep -qFe ",$Val," <<<",$Ref,";;
        veq)                __dist_vercmp_eq          ;;
        vne)                ! __dist_vercmp_eq        ;;
        vge|vle|vgt|vlt)    __dist_vercmp             ;;
        *) mkusage -w "unsupported OP: $Op" \
                   -E "KEY OP VALUE [KEY OP VALUE].."
           return 2 ;;
    esac
}

__dist_intcmp() {
    #
    # Compare integers $Val and $Ref with $Op
    #
    isa__int "$Val" || { warn "not an integer: '$Val'"; return 3; }
    isa__int "$Ref" || { warn "not an integer: '$Ref'"; return 3; }
    test "$Val" -"$Op" "$Ref"
}

__dist_vercmp_eq() {
    #
    # Decide version match/equality
    #
    local val_f=()  # $Val fragments
    local ref_f=()  # $Ref fragments
    local idx=0     # compared index
    IFS=. read -ra val_f <<<"$Val"
    IFS=. read -ra ref_f <<<"$Ref"
    test -n "${val_f[0]}" || { warn "invalid version: $Val"; return 3; }
    test -n "${ref_f[0]}" || { warn "invalid version: $Ref"; return 3; }
    while true; do                          # Val='1.2.3'  Ref='1.2'
        v=${val_f[$idx]}; r=${ref_f[$idx]}  # v=1 r=1 | v=2 r=2 | v=3 r=
        test -n "$v" || return 0            # one of them ran out
        test -n "$r" || return 0            # ..
        test "$v" -eq "$r" || return 1      # one of them differs
        ((idx++))
    done
}

__dist_vercmp() {
    #
    # Compare versions $Val and $Ref with $Op
    #
    local rv="$Ref"$'\n'"$Val"  # $Ref then $Val
    local vr="$Val"$'\n'"$Ref"  # $Val then $Ref
    local rvs                   # ^^ above, but version-sorted
    local vrs                   # ^^ above, but version-sorted
    local rvs_same=0            # 1 if $rv respoded to sorting, 0 otherwise
    local vrs_same=0            # 1 if $vr respoded to sorting, 0 otherwise
    rvs=$(sort -V <<<"$rv")
    vrs=$(sort -V <<<"$vr")
    test "$rvs" == "$rv"; rvs_same=$?   # 0: r is ge    1: r is lt
    test "$vrs" == "$vr"; vrs_same=$?   # 0: v is ge    1: v is lt
    case ${Op:1}:$rvs_same:$vrs_same in

        # both are same
        #
        eq:0:0)  return 0 ;;
        ge:0:0)  return 0 ;;
        le:0:0)  return 0 ;;
        gt:0:0)  return 1 ;;
        lt:0:0)  return 1 ;;
        ne:0:0)  return 1 ;;

        # r is ge, v is lt
        #
        eq:0:1)  return 1 ;;
        ge:0:1)  return 0 ;;
        gt:0:1)  return 0 ;;
        le:0:1)  return 0 ;;
        lt:0:1)  return 1 ;;
        ne:0:1)  return 0 ;;

        # r is lt, v is ge
        #
        eq:1:0)  return 1 ;;
        ge:1:0)  return 1 ;;
        gt:1:0)  return 1 ;;
        le:1:0)  return 1 ;;
        lt:1:0)  return 0 ;;
        ne:1:0)  return 0 ;;

    esac
    #FIXME: these are pretty much surely borken
}

#shellfu module-version=0.2.5
